(*
 * Varna - sister of Red Sonja
 * 
 * this programm permutes a given string set
 *
 * in opposite to Red Sonja it does not use constraints
 *
 * (c) 2008 by Andreas Romeyke
 * distributed under terms of GPL v3, 
 * see License file "gpl-3.0.txt"
 * or http://www.gnu.org/licenses/gpl-3.0.txt
 *
 *)

open ExtLib

let print_help () = 
  Printf.eprintf "\ncalcs the words of given length with symbols of given strings\n";
  Printf.eprintf "call:\n";
  Printf.eprintf "\tvarna <wordlen> <alpha> [<beta> .. <omega>]\n";
;;

let remove_dublettes words =
  let whash = Hashtbl.create 0 in
    List.filter (
      fun w -> try
        Hashtbl.find whash w;
        false
      with
          Not_found ->
            begin
              Hashtbl.add whash w true;
              true
            end
    ) words
;;

(* variant derived from Andreas old variant by William D. Neuman,
 * optimized by Christoph Troest to use less memory
 * adapted to varna (without constraints)
 *)
let combine_3 ls1 ls2 = 
  let comb l1 l2 =
    let add_prod_l1 acc v2 = 
      List.fold_left (fun acc v1 ->
                        let v12 = v1 @ v2 in
                          v12::acc
      ) acc l1
    in
      List.fold_left add_prod_l1 [] l2
  in
    comb ls1 ls2
;;

(* permute at n *)
let permute (alpha: string list) wordlen =
  (* object hash -> maps strings to id to reduce overhead in recursion *)
  let objhash = Hashtbl.create (List.length alpha) in
  let inv_objhash = Hashtbl.create (List.length alpha) in
  (* map objects to ids *)
  let (alpha:int list) = List.map (
    fun v ->
      let (idx,freq) = 
        try
          Hashtbl.find inv_objhash v
        with
            Not_found -> (Hashtbl.length objhash, 0) 
      in
        Hashtbl.replace objhash idx (v,freq+1);
        Hashtbl.replace inv_objhash v (idx,freq+1);
        idx
  ) alpha in
  let alpha = remove_dublettes alpha 
  in
    Sys.catch_break true;
    Printf.eprintf "Varna - Sister of Red Sonja\n";
    Printf.eprintf "(c) Andreas Romeyke (art1@andreas-romeyke.de)\n";
    Printf.eprintf "    distributed under terms of GPL v3 or higher\n";
    Printf.eprintf "    see http://www.gnu.org/licenses/gpl-3.0.txt\n";
    Printf.eprintf "    or License file \"gpl-3.0.txt\"\n";
    print_help ();
    Printf.eprintf "---------------------------------------------\n";
    let abc =  (List.length alpha) in
    let mpermut = ((float_of_int abc) ** (float_of_int wordlen)) in
      if mpermut > 5514439240800.0 then 
        begin
          Printf.eprintf "The calc of powerset with %.0f permutations needs about 1 Month, aborting\n" mpermut;
          raise Exit
        end;
      Printf.eprintf "wordlen: %i alphabet: %i  permutations: %e\n"  wordlen abc mpermut  ; 
      Printf.eprintf "++-----------++----------++----------++--------++--------------++\n";
      Printf.eprintf "|| recursion ||   words  ||    rtime ||    %%   ||   words/s    ||\n";
      Printf.eprintf "++-----------++----------++----------++--------++--------------++\n";
      let alphabet = List.map (fun x -> [x]) alpha in
      let rec permute_r (n: int) = 
        match n with
            0 -> [] 
          | 1 -> alphabet
          | _  ->  
              let a = n/2
              in
              let b = n-a
              in
              let words = 
                let perm_a = permute_r a in
                  if a == b then 
                    combine_3 (perm_a) (perm_a)
                  else
                    combine_3 (perm_a) (permute_r b)
              in 
              let wlen = (List.length words) in
              let rtime = Sys.time() in
                Printf.eprintf "\r||  %8i || %3.2e || %08.2f ||%7.3f || %12.2f ||     " n (float_of_int wlen) rtime (100.0 /. mpermut *. (float_of_int wlen)) ((float_of_int wlen) /. rtime);
                flush_all ();
                words
      in
        try
          (* map id to obj *)
          let (pot:int list list) = permute_r wordlen in
            List.map (
              List.map (
                fun x ->
                  let (v,f) = 
                    Hashtbl.find objhash x
                  in
                    v
              )
            ) pot
        with
            Sys.Break -> 
              Printf.eprintf "\n... Aborting...\n";
              [[]]
;;


(** MAIN **)
try
  let lst = List.tl (Array.to_list Sys.argv) in
    if List.length lst < 2 then 
      begin
        Printf.eprintf "missed arguments!\n";
        print_help ();
        raise Exit

      end;
    let (len:int) = int_of_string (List.hd lst) in
    let alphabet = List.tl lst in
      List.iter (
        fun sl ->
          List.iter (
            Printf.printf "%s "
          ) sl; 
          Printf.printf "\n";

      ) (permute alphabet len);
      Printf.eprintf "\n"
with Exit -> ()
;;



